@@ -46,6 +46,14 @@ const pfmlib_attr_desc_t snbep_unc_mods[]={
 	PFM_ATTR_B("isoc", "match isochronous requests"),   /* isochronous */
 	PFM_ATTR_B("nc", "match non-coherent requests"),   /* non-coherent */
 	PFM_ATTR_I("cf", "core id filter, includes non-thread data in bit 5 [0-63]"),	/* core id (hswep) */
+	PFM_ATTR_I("tf", "thread id filter [0-3]"),	 /* thread id (skx)*/
+	PFM_ATTR_I("cf", "source id filter [0-63]"),	 /* src-id/core-id (skx) */
+	PFM_ATTR_B("loc", "match on local node target"), /* loc filter1 (skx) */
+	PFM_ATTR_B("rem", "match on remote node target"),/* rem filter1 (skx) */
+	PFM_ATTR_B("lmem", "local memory cacheable"),	 /* nm filter1 (skx) */
+	PFM_ATTR_B("rmem", "remote memory cacheable"),	 /* not_nm filter1 (skx) */
+	PFM_ATTR_I("dnid", "destination node id [0-15]"), /* SKX:UPI */
+	PFM_ATTR_I("rcsnid", "destination RCS Node id [0-15]"), /* SKX:UPI */
 	PFM_ATTR_NULL
 };
 
@@ -121,7 +129,27 @@ pfm_intel_knl_unc_detect(void *this)
                return PFM_ERR_NOTSUPP;
 
        switch(pfm_intel_x86_cfg.model) {
-               case 87: /* Knights Landing */
+               case 87:  /* Knights Landing */
+                         break;
+               default:
+                       return PFM_ERR_NOTSUPP;
+       }
+       return PFM_SUCCESS;
+}
+
+int
+pfm_intel_knm_unc_detect(void *this)
+{
+       int ret;
+
+       ret = pfm_intel_x86_detect();
+       if (ret != PFM_SUCCESS)
+
+       if (pfm_intel_x86_cfg.family != 6)
+               return PFM_ERR_NOTSUPP;
+
+       switch(pfm_intel_x86_cfg.model) {
+               case 133: /* Knights Mill */
                          break;
                default:
                        return PFM_ERR_NOTSUPP;
@@ -150,6 +178,27 @@ pfm_intel_bdx_unc_detect(void *this)
        return PFM_SUCCESS;
 }
 
+int
+pfm_intel_skx_unc_detect(void *this)
+{
+       int ret;
+
+       ret = pfm_intel_x86_detect();
+       if (ret != PFM_SUCCESS)
+
+       if (pfm_intel_x86_cfg.family != 6)
+               return PFM_ERR_NOTSUPP;
+
+       switch(pfm_intel_x86_cfg.model) {
+               case 85: /* Skylake X */
+                         break;
+               default:
+                       return PFM_ERR_NOTSUPP;
+       }
+       return PFM_SUCCESS;
+}
+
+
 static void
 display_com(void *this, pfmlib_event_desc_t *e, void *val)
 {
@@ -194,12 +243,25 @@ get_pcu_filt_band(void *this, pfm_snbep_unc_reg_t reg)
 	return reg.pcu.unc_event - PCU_FREQ_BAND0_CODE;
 }
 
+static inline void
+set_filters(void *this, pfm_snbep_unc_reg_t *filters, int event, int umask)
+{
+	const intel_x86_entry_t *pe = this_pe(this);
+
+	filters[0].val |= pe[event].umasks[umask].ufilters[0] & ((1ULL << 32)-1);
+	filters[0].val &= ~(pe[event].umasks[umask].ufilters[0] >> 32);
+
+	filters[1].val |= pe[event].umasks[umask].ufilters[1] & ((1ULL << 32)-1);
+	filters[1].val &= ~(pe[event].umasks[umask].ufilters[1] >>32);
+}
+
 int
 snbep_unc_add_defaults(void *this, pfmlib_event_desc_t *e,
 			   unsigned int msk,
 			   uint64_t *umask,
 			   pfm_snbep_unc_reg_t *filters,
-			   unsigned short max_grpid)
+			   unsigned short max_grpid,
+			   int *numasks)
 {
 	const intel_x86_entry_t *pe = this_pe(this);
 	const intel_x86_entry_t *ent;
@@ -226,7 +288,7 @@ snbep_unc_add_defaults(void *this, pfmlib_event_desc_t *e,
 
 			idx = e->pattrs[j].idx;
 
-			if (ent->umasks[idx].grpid != i)
+			if (get_grpid(ent->umasks[idx].grpid) != i)
 				continue;
 
 			if (max_grpid != INTEL_X86_MAX_GRPID && i > max_grpid) {
@@ -253,20 +315,20 @@ snbep_unc_add_defaults(void *this, pfmlib_event_desc_t *e,
 				 */
 				*umask |= ent->umasks[idx].ucode >> 8;
 
-				filters[0].val |= pe[e->event].umasks[idx].ufilters[0];
-				filters[1].val |= pe[e->event].umasks[idx].ufilters[1];
+				set_filters(this, filters, e->event, idx);
 
 				e->attrs[k].id = j; /* pattrs index */
 				e->attrs[k].ival = 0;
 				k++;
 
+				(*numasks)++;
 				added++;
 				if (intel_x86_eflag(this, e->event, INTEL_X86_GRP_EXCL))
 					goto done;
 
 				if (intel_x86_uflag(this, e->event, idx, INTEL_X86_EXCL_GRP_GT)) {
 					if (max_grpid != INTEL_X86_MAX_GRPID) {
-						DPRINT("two max_grpid, old=%d new=%d\n", max_grpid, ent->umasks[idx].grpid);
+						DPRINT("two max_grpid, old=%d new=%d\n", max_grpid, get_grpid(ent->umasks[idx].grpid));
 						return PFM_ERR_UMASK;
 					}
 					max_grpid = ent->umasks[idx].grpid;
@@ -295,6 +357,7 @@ pfm_intel_snbep_unc_get_encoding(void *this, pfmlib_event_desc_t *e)
 	unsigned int grpmsk, ugrpmsk = 0;
 	unsigned short max_grpid = INTEL_X86_MAX_GRPID;
 	unsigned short last_grpid =  INTEL_X86_MAX_GRPID;
+	unsigned short req_grpid;
 	int umodmsk = 0, modmsk_r = 0;
 	int pcu_filt_band = -1;
 	pfm_snbep_unc_reg_t reg;
@@ -302,10 +365,12 @@ pfm_intel_snbep_unc_get_encoding(void *this, pfmlib_event_desc_t *e)
 	pfm_snbep_unc_reg_t addr;
 	pfmlib_event_attr_info_t *a;
 	uint64_t val, umask1, umask2;
-	int k, ret;
-	int has_cbo_tid = 0;
+	int k, ret, numasks = 0;
+	int must_have_filt0 = 0;
+	int max_req_grpid = -1;
 	unsigned short grpid;
 	int grpcounts[INTEL_X86_NUM_GRP];
+	int req_grps[INTEL_X86_NUM_GRP];
 	int ncombo[INTEL_X86_NUM_GRP];
 	char umask_str[PFMLIB_EVT_MAX_NAME_LEN];
 
@@ -315,20 +380,21 @@ pfm_intel_snbep_unc_get_encoding(void *this, pfmlib_event_desc_t *e)
 
 	addr.val = 0;
 
-	pe = this_pe(this);
-
 	umask_str[0] = e->fstr[0] = '\0';
 
 	reg.val = val = pe[e->event].code;
 
 	/* take into account hardcoded umask */
-	umask1 = (val >> 8) & 0xff;
+	umask1 = (val >> 8);
 	umask2 = umask1;
 
 	grpmsk = (1 << pe[e->event].ngrp)-1;
 
 	modmsk_r = pe[e->event].modmsk_req;
 
+	if (intel_x86_eflag(this, e->event, INTEL_X86_FORCE_FILT0))
+		must_have_filt0 = 1;
+
 	for(k=0; k < e->nattrs; k++) {
 		a = attr(e, k);
 
@@ -338,7 +404,8 @@ pfm_intel_snbep_unc_get_encoding(void *this, pfmlib_event_desc_t *e)
 		if (a->type == PFM_ATTR_UMASK) {
 			uint64_t um;
 
-			grpid = pe[e->event].umasks[a->idx].grpid;
+			grpid     = get_grpid(pe[e->event].umasks[a->idx].grpid);
+			req_grpid = get_req_grpid(pe[e->event].umasks[a->idx].grpid);
 
 			/*
 			 * certain event groups are meant to be
@@ -379,6 +446,19 @@ pfm_intel_snbep_unc_get_encoding(void *this, pfmlib_event_desc_t *e)
 			 * umask group
 			 */
 			++grpcounts[grpid];
+			if (intel_x86_uflag(this, e->event, a->idx, INTEL_X86_GRP_REQ)) {
+				DPRINT("event requires grpid %d\n", req_grpid);
+				/* initialize req_grpcounts array only when needed */
+				if (max_req_grpid == -1) {
+					int x;
+					for (x = 0; x < INTEL_X86_NUM_GRP; x++)
+						req_grps[x] = 0xff;
+				}
+				if (req_grpid > max_req_grpid)
+					max_req_grpid = req_grpid;
+				DPRINT("max_req_grpid=%d\n", max_req_grpid);
+				req_grps[req_grpid] = 1;
+			}
 
 			/* mark that we have a umask with NCOMBO in this group */
 			if (intel_x86_uflag(this, e->event, a->idx, INTEL_X86_NCOMBO))
@@ -397,8 +477,8 @@ pfm_intel_snbep_unc_get_encoding(void *this, pfmlib_event_desc_t *e)
 			last_grpid = grpid;
 
 			um = pe[e->event].umasks[a->idx].ucode;
-			filters[0].val |= pe[e->event].umasks[a->idx].ufilters[0];
-			filters[1].val |= pe[e->event].umasks[a->idx].ufilters[1];
+
+			set_filters(this, filters, e->event, a->idx);
 
 			um >>= 8;
 			umask2  |= um;
@@ -415,6 +495,7 @@ pfm_intel_snbep_unc_get_encoding(void *this, pfmlib_event_desc_t *e)
 			evt_strcat(umask_str, ":%s", pe[e->event].umasks[a->idx].uname);
 
 			modmsk_r |= pe[e->event].umasks[a->idx].umodmsk_req;
+			numasks++;
 
 		} else if (a->type == PFM_ATTR_RAW_UMASK) {
 
@@ -428,6 +509,7 @@ pfm_intel_snbep_unc_get_encoding(void *this, pfmlib_event_desc_t *e)
 			/* override umask */
 			umask2 = a->idx & 0xff;
 			ugrpmsk = grpmsk;
+			numasks++;
 		} else {
 			uint64_t ival = e->attrs[k].ival;
 			switch(a->idx) {
@@ -465,16 +547,24 @@ pfm_intel_snbep_unc_get_encoding(void *this, pfmlib_event_desc_t *e)
 						return PFM_ERR_ATTR_VAL;
 					}
 					reg.cbo.unc_tid = 1;
-					has_cbo_tid = 1;
+					must_have_filt0 = 1;
 					filters[0].cbo_filt.tid = ival;
 					umodmsk |= _SNBEP_UNC_ATTR_TF;
 					break;
+				case SNBEP_UNC_ATTR_TF1: /* thread id skx */
+					if (ival > 7)
+						return PFM_ERR_ATTR_VAL;
+					reg.cha.unc_tid = 1;
+					filters[0].skx_cha_filt0.tid = ival; /* includes non-thread data */
+					must_have_filt0 = 1;
+					umodmsk |= _SNBEP_UNC_ATTR_TF1;
+					break;
 				case SNBEP_UNC_ATTR_CF: /* core id */
 					if (ival > 15)
 						return PFM_ERR_ATTR_VAL;
 					reg.cbo.unc_tid = 1;
 					filters[0].cbo_filt.cid = ival;
-					has_cbo_tid = 1;
+					must_have_filt0 = 1;
 					umodmsk |= _SNBEP_UNC_ATTR_CF;
 					break;
 				case SNBEP_UNC_ATTR_CF1: /* core id */
@@ -482,7 +572,7 @@ pfm_intel_snbep_unc_get_encoding(void *this, pfmlib_event_desc_t *e)
 						return PFM_ERR_ATTR_VAL;
 					reg.cbo.unc_tid = 1;
 					filters[0].hswep_cbo_filt0.cid = ival; /* includes non-thread data */
-					has_cbo_tid = 1;
+					must_have_filt0 = 1;
 					umodmsk |= _SNBEP_UNC_ATTR_CF1;
 					break;
 				case SNBEP_UNC_ATTR_NF: /* node id filter0 */
@@ -501,6 +591,14 @@ pfm_intel_snbep_unc_get_encoding(void *this, pfmlib_event_desc_t *e)
 					filters[1].ivbep_cbo_filt1.nid = ival;
 					umodmsk |= _SNBEP_UNC_ATTR_NF1;
 					break;
+				case SNBEP_UNC_ATTR_CF2: /* src-id/core-id skx */
+					if (ival > 64)
+						return PFM_ERR_ATTR_VAL;
+					reg.cha.unc_tid = 1;
+					filters[0].skx_cha_filt0.sid = ival;
+					must_have_filt0 = 1;
+					umodmsk |= _SNBEP_UNC_ATTR_CF2;
+					break;
 				case SNBEP_UNC_ATTR_FF: /* freq band filter */
 					if (ival > 255)
 						return PFM_ERR_ATTR_VAL;
@@ -523,20 +621,84 @@ pfm_intel_snbep_unc_get_encoding(void *this, pfmlib_event_desc_t *e)
 				case SNBEP_UNC_ATTR_NC: /* nc filter */
 					filters[1].ivbep_cbo_filt1.nc = !!ival;
 					break;
+				case SNBEP_UNC_ATTR_LOC: /* local target skx */
+					filters[1].skx_cha_filt1.loc = !!ival;
+					break;
+				case SNBEP_UNC_ATTR_REM: /* remote target skx */
+					filters[1].skx_cha_filt1.rem = !!ival;
+					break;
+				case SNBEP_UNC_ATTR_LMEM: /* local target skx */
+					filters[1].skx_cha_filt1.loc = !!ival;
+					break;
+				case SNBEP_UNC_ATTR_RMEM: /* local memory skx */
+					filters[1].skx_cha_filt1.not_nm = !!ival;
+					break;
+				case SNBEP_UNC_ATTR_DNID: /* destination node id skx */
+					if (ival > 15) {
+						DPRINT("dnid must be [0-15]\n");
+						return PFM_ERR_ATTR_VAL;
+					}
+					filters[0].skx_upi_filt.dnid = ival;
+					filters[0].skx_upi_filt.en_dnidd = 1;
+					break;
+				case SNBEP_UNC_ATTR_RCSNID: /* RCS node id skx */
+					if (ival > 15) {
+						DPRINT("rcsnid must be [0-15]\n");
+						return PFM_ERR_ATTR_VAL;
+					}
+					filters[0].skx_upi_filt.rcsnid = ival;
+					filters[0].skx_upi_filt.en_rcsnid = 1;
+					break;
+				default:
+					DPRINT("event %s invalid attribute %d\n", pe[e->event].name, a->idx);
+					return PFM_ERR_ATTR;
+					
 			}
 		}
 	}
+
+	/* check required groups are in place */
+	if (max_req_grpid != -1) {
+		int x;
+		for (x = 0; x <= max_req_grpid; x++) {
+			if (req_grps[x] == 0xff)
+				continue;
+			if ((ugrpmsk & (1 << x)) == 0) {
+				DPRINT("required grpid %d umask missing\n", x);
+				return PFM_ERR_FEATCOMB;
+			}
+		}
+	}
+
+	/* check required groups are in place */
+	if (max_req_grpid != -1) {
+		int x;
+		for (x = 0; x <= max_req_grpid; x++) {
+			if (req_grps[x] == 0xff)
+				continue;
+			if ((ugrpmsk & (1 << x)) == 0) {
+				DPRINT("required grpid %d umask missing\n", x);
+				return PFM_ERR_FEATCOMB;
+			}
+		}
+	}
+
 	/*
 	 * check that there is at least of unit mask in each unit mask group
 	 */
 	if (pe[e->event].numasks && (ugrpmsk != grpmsk || ugrpmsk == 0)) {
 		uint64_t um = 0;
 		ugrpmsk ^= grpmsk;
-		ret = snbep_unc_add_defaults(this, e, ugrpmsk, &um, filters, max_grpid);
+		ret = snbep_unc_add_defaults(this, e, ugrpmsk, &um, filters, max_grpid, &numasks);
 		if (ret != PFM_SUCCESS)
 			return ret;
 		umask2 |= um;
 	}
+	/* if event has umasks, then likely at least one must be set */
+	if (pe[e->event].numasks && numasks == 0) {
+		DPRINT("event has umasks but none specified\n");
+		return PFM_ERR_ATTR;
+	}
 
 	/*
 	 * nf= is only required on some events in CBO
@@ -555,6 +717,19 @@ pfm_intel_snbep_unc_get_encoding(void *this, pfmlib_event_desc_t *e)
 		return PFM_ERR_ATTR;
 	}
 
+	/*
+	 * fixup filt1.all_opc based on values of the filter
+	 */
+	if (is_cha_filt_event(this, 1, reg)) {
+		if (filters[1].val == 0)
+			/* default value: rem=loc=nm=not_nm=all_opc=1 */
+			filters[1].val =0x3b;
+		else if (filters[1].skx_cha_filt1.opc0 || filters[1].skx_cha_filt1.opc1) {
+			/* enable opcode filtering */
+			filters[1].val &= ~(1ULL << 3);
+		}
+	}
+
 	evt_strcat(e->fstr, "%s", pe[e->event].name);
 	pfmlib_sort_attr(e);
 
@@ -574,9 +749,9 @@ pfm_intel_snbep_unc_get_encoding(void *this, pfmlib_event_desc_t *e)
 	e->codes[e->count++] = reg.val;
 
 	/*
-	 * handles C-box filter
+	 * handles filters
 	 */
-	if (filters[0].val || filters[1].val || has_cbo_tid)
+	if (filters[0].val || filters[1].val || must_have_filt0)
 		e->codes[e->count++] = filters[0].val;
 	if (filters[1].val)
 		e->codes[e->count++] = filters[1].val;
@@ -617,12 +792,18 @@ pfm_intel_snbep_unc_get_encoding(void *this, pfmlib_event_desc_t *e)
 		case SNBEP_UNC_ATTR_TF:
 			evt_strcat(e->fstr, ":%s=%lu", snbep_unc_mods[idx].name, reg.cbo.unc_tid);
 			break;
+		case SNBEP_UNC_ATTR_TF1:
+			evt_strcat(e->fstr, ":%s=%lu", snbep_unc_mods[idx].name, filters[0].skx_cha_filt0.tid);
+			break;
 		case SNBEP_UNC_ATTR_CF:
 			evt_strcat(e->fstr, ":%s=%lu", snbep_unc_mods[idx].name, filters[0].cbo_filt.cid);
 			break;
 		case SNBEP_UNC_ATTR_CF1:
 			evt_strcat(e->fstr, ":%s=%lu", snbep_unc_mods[idx].name, filters[0].hswep_cbo_filt0.cid);
 			break;
+		case SNBEP_UNC_ATTR_CF2:
+			evt_strcat(e->fstr, ":%s=%lu", snbep_unc_mods[idx].name, filters[0].skx_cha_filt0.sid);
+			break;
 		case SNBEP_UNC_ATTR_FF:
 			evt_strcat(e->fstr, ":%s=%lu", snbep_unc_mods[idx].name, (filters[0].val >> (pcu_filt_band*8)) & 0xff);
 			break;
@@ -644,6 +825,27 @@ pfm_intel_snbep_unc_get_encoding(void *this, pfmlib_event_desc_t *e)
 			evt_strcat(e->fstr, ":%s=0x%lx", snbep_unc_mods[idx].name,
 				   addr.ha_addr.hi_addr << 26 | addr.ha_addr.lo_addr);
 			break;
+		case SNBEP_UNC_ATTR_REM:
+			evt_strcat(e->fstr, ":%s=%lu", snbep_unc_mods[idx].name, filters[1].skx_cha_filt1.rem);
+			break;
+		case SNBEP_UNC_ATTR_LOC:
+			evt_strcat(e->fstr, ":%s=%lu", snbep_unc_mods[idx].name, filters[1].skx_cha_filt1.loc);
+			break;
+		case SNBEP_UNC_ATTR_RMEM:
+			evt_strcat(e->fstr, ":%s=%lu", snbep_unc_mods[idx].name, filters[1].skx_cha_filt1.rem);
+			break;
+		case SNBEP_UNC_ATTR_LMEM:
+			evt_strcat(e->fstr, ":%s=%lu", snbep_unc_mods[idx].name, filters[1].skx_cha_filt1.loc);
+			break;
+		case SNBEP_UNC_ATTR_DNID:
+			evt_strcat(e->fstr, ":%s=%lu", snbep_unc_mods[idx].name, filters[0].skx_upi_filt.dnid);
+			break;
+		case SNBEP_UNC_ATTR_RCSNID:
+			evt_strcat(e->fstr, ":%s=%lu", snbep_unc_mods[idx].name, filters[0].skx_upi_filt.rcsnid);
+			break;
+		default:
+			DPRINT("unknown attribute %d for event %s\n", idx, pe[e->event].name);
+			return PFM_ERR_ATTR;
 		}
 	}
 	display_reg(this, e, reg);
